import sqlalchemy as db
from importlib import import_module
import envtoml


def prepare_table(engine, table_name):
    metadata = db.MetaData()
    return db.Table(table_name, metadata, autoload=True, autoload_with=engine)


def execute_validation(engine, table, column, validator, args=None,
                       custom_column=None):
    conn = engine.connect()
    t = prepare_table(engine, table)
    col = t.columns[column]

    # If a custom column has been specified, use it in place of the column from
    # the table
    if custom_column:
        col = db.sql.literal_column(custom_column)

    s = validator(t, col, args=args)
    ex = conn.execute(s)

    if type(s) == str:
        bare_query = s
    else:
        compile_kwargs = {"literal_binds": True}
        bare_query = s.compile(engine, compile_kwargs=compile_kwargs)

    return ex.fetchall(), ex.keys(), bare_query


def execute_validations(config):
    cfg = envtoml.load(config)

    engine = db.create_engine(cfg['general']['sqla'])
    db_name = cfg['general']['db_name']

    validator_module = import_module('.validators', package='sqvid')

    for table in cfg[db_name]:
        for column in cfg[db_name][table]:
            for val in cfg[db_name][table][column]:
                validator_name = val['validator']
                args = val.get('args')
                custom_column = val.get('custom_column')
                severity = val.get('severity', 'error')

                col_name = column
                if custom_column:
                    col_name = "{} (customized as '{}')".format(column,
                                                                custom_column)
                result = None
                further_info = ''

                validator_fn = getattr(validator_module, validator_name)
                try:
                    r, k, q = execute_validation(engine, table, column,
                                                 validator_fn, args,
                                                 custom_column=custom_column)
                except KeyError as e:
                    further_info = " - {} could not be found".format(e)
                except Exception as e:
                    further_info = " - {}: {}".format(e.__class__.__name__, e)

                if further_info:
                    # since the validation errored out there are no columns to
                    # be reported and the query is also not really relevant
                    k = []
                    r = []
                    q = '-- ERROR --'

                    result = 'error'

                col_names = val.get('report_columns', k)

                if not result:
                    if len(r) == 0:
                        result = 'passed'
                    elif severity == 'warn':
                        result = 'failed (warn only)'
                    else:
                        result = 'failed'

                out = "{}: Validation on [{}] {}.{} of {}{}{}".format(
                    result.upper(),
                    db_name,
                    table,
                    col_name,
                    validator_name,
                    '({})'.format(args) if args else '',
                    further_info
                )

                yield {
                    'query': q,
                    'result': result,
                    'output': out,
                    'rows': list(map(dict, r)),
                    'col_names': col_names
                }



from .executor import execute_validations
from nicetable.nicetable import NiceTable
import sys
import click

QUERY_VERBOSE_STR = "\nRUNNING QUERY:\n===========\n{}\n==========="


@click.command()
@click.option('--config',
              type=click.Path(exists=True),
              required=True,
              help='Path to a .toml config file.')
@click.option('--verbose/--no-verbose', default=False,
              help='Turn on verbose output of SQL queries.')
def run(config, verbose):
    """Validator of data that is queriable via SQL."""

    n_failed = 0
    for validation in execute_validations(config):
        if verbose:
            print(QUERY_VERBOSE_STR.format(validation['query']))

        print(validation['output'])

        if validation['result'] in ['failed', 'error']:
            n_failed += 1

        if validation['result'].startswith('failed'):
            print("Offending {} rows:".format(len(validation['rows'])))
            print(NiceTable(validation['rows'],
                            col_names=validation['col_names']))

    if n_failed > 0:
        sys.exit(1)
    else:
        sys.exit(0)



import sqlalchemy as db
from jinjasql import JinjaSql


def in_range(table, column, args=None):
    """
    Check whether values in a column fall within a specific range.

    Args:
        min (int): the minimum value of the range
        max (int): the maximum value of the range

    Example:
      .. code:: toml

        [[test_sqvid_db.suppliers.SupplierID]]
        validator = 'in_range'
        args = {min = 1, max = 256}

    """
    assert 'min' in args
    assert 'max' in args

    return db.select([table]).where(db.or_(column < args['min'],
                                           column > args['max']))


def accepted_values(table, column, args=None):
    """
    Check that a column contains only specified values.

    Args:
        vals (list): a list of values

    Example:
      .. code:: toml

        [[test_sqvid_db.suppliers.Country]]
        validator = 'accepted_values'
        args.vals  = [
          'USA',
          'UK',
          'Spain',
          'Japan',
          'Germany',
          'Australia',
          'Sweden',
          'Finland',
          'Italy',
          'Brazil',
          'Singapore',
          'Norway',
          'Canada',
          'France',
          'Denmark',
          'Netherlands'
        ]


    """
    assert 'vals' in args

    return db.select([table]).where(column.notin_(args['vals']))


def not_null(table, column, args=None):
    """
    Check that a column contains only non-`NULL` values.

    Example:
      .. code:: toml

        [[test_sqvid_db.suppliers.SupplierID]]
        validator = 'not_null'
    """
    return db.select([table]).where(column.is_(None))


def unique(table, column, args=None):
    """
    Check whether values in a column are unique.

    Example:
      .. code:: toml

        [[test_sqvid_db.suppliers.SupplierID]]
        validator = 'unique'
    """
    return db.select([column]) \
        .select_from(table) \
        .group_by(column) \
        .having(db.func.count(column) > 1)


def custom_sql(table, column, args=None):
    """
    Execute a custom (optionally Jinja-formatted) SQL query and fail if
    non-zero number of rows is returned.

    Either ``query`` or ``query_file`` parameter needs to be provided. All the
    other arguments are passed as Jinja variables and can be used to build the
    query.

    Args:
        query (str): query to be executed (optional).
        query_file (str): path to the file in which the query to be executed
                          can be found (optional)

    Example:
      .. code:: toml

        [[test_sqvid_db.suppliers.SupplierID]]
        validator = 'custom_sql'
        args.query_file = './tests/queries/tables_equal_rows.sql'
        args.other_table = 'suppliers_copy'
    """
    j = JinjaSql(param_style='pyformat')

    if 'query_file' in args:
        q_str = open(args['query_file']).read()
    if 'query' in args:
        q_str = args['query']

    args['table'] = table.name
    args['column'] = column.name

    query, bind_params = j.prepare_query(q_str, args)

    return query % bind_params
